import re
import string
import typing

import pytest
from dagster import (
    Any,
    DagsterInvalidConfigDefinitionError,
    DagsterInvalidConfigError,
    DagsterInvalidDefinitionError,
    Field,
    Float,
    Int,
    List,
    ModeDefinition,
    Noneable,
    Permissive,
    PipelineDefinition,
    ResourceDefinition,
    Set,
    String,
    Tuple,
    composite_solid,
    execute_pipeline,
    execute_solid,
    pipeline,
    solid,
)
from dagster.config.errors import DagsterEvaluationErrorReason
from dagster.config.field_utils import convert_potential_field
from dagster.config.validate import process_config, validate_config


def test_noop_config():
    assert Field(Any)


def test_int_field():
    config_field = convert_potential_field({"int_field": Int})
    assert validate_config(config_field.config_type, {"int_field": 1}).value == {"int_field": 1}


def test_float_field():
    config_field = convert_potential_field({"float_field": Float})
    assert validate_config(config_field.config_type, {"float_field": 1.0}).value == {
        "float_field": 1.0
    }
    assert process_config(config_field.config_type, {"float_field": 1.0}).value == {
        "float_field": 1.0
    }
    assert validate_config(config_field.config_type, {"float_field": 1}).value == {"float_field": 1}
    assert process_config(config_field.config_type, {"float_field": 1}).value == {
        "float_field": 1.0
    }


def assert_config_value_success(config_type, config_value, expected):
    result = process_config(config_type, config_value)
    assert result.success
    assert result.value == expected


def assert_eval_failure(config_type, value):
    assert not validate_config(config_type, value).success


def test_int_fails():
    config_field = convert_potential_field({"int_field": Int})

    assert_eval_failure(config_field.config_type, {"int_field": "fjkdj"})
    assert_eval_failure(config_field.config_type, {"int_field": True})


def test_default_arg():
    config_field = convert_potential_field(
        {"int_field": Field(Int, default_value=2, is_required=False)}
    )

    assert_config_value_success(config_field.config_type, {}, {"int_field": 2})


def test_default_float_arg():
    config_field = convert_potential_field(
        {"float_field": Field(Float, default_value=2.0, is_required=False)}
    )

    assert_config_value_success(config_field.config_type, {}, {"float_field": 2.0})

    config_field = convert_potential_field(
        {"float_field": Field(Float, default_value=2, is_required=False)}
    )

    assert_config_value_success(config_field.config_type, {}, {"float_field": 2})


def _single_required_string_config_dict():
    return convert_potential_field({"string_field": String})


def _multiple_required_fields_config_dict():
    return convert_potential_field({"field_one": String, "field_two": String})


def _single_optional_string_config_dict():
    return convert_potential_field({"optional_field": Field(String, is_required=False)})


def _single_optional_string_field_config_dict_with_default():
    optional_field_def = Field(String, is_required=False, default_value="some_default")
    return convert_potential_field({"optional_field": optional_field_def})


def _mixed_required_optional_string_config_dict_with_default():
    return convert_potential_field(
        {
            "optional_arg": Field(String, is_required=False, default_value="some_default"),
            "required_arg": Field(String, is_required=True),
            "optional_arg_no_default": Field(String, is_required=False),
        }
    )


def _multiple_required_fields_config_permissive_dict():
    return Field(Permissive({"field_one": Field(String), "field_two": Field(String)}))


def _validate(config_field, value):
    res = process_config(config_field.config_type, value)
    assert res.success, res.errors[0].message
    return res.value


def test_single_required_string_field_config_type():
    assert _validate(_single_required_string_config_dict(), {"string_field": "value"}) == {
        "string_field": "value"
    }

    with pytest.raises(
        AssertionError,
        match='Missing required config entry "string_field" at the root.',
    ):
        _validate(_single_required_string_config_dict(), {})

    with pytest.raises(AssertionError):
        _validate(_single_required_string_config_dict(), {"extra": "yup"})

    with pytest.raises(AssertionError):
        _validate(_single_required_string_config_dict(), {"string_field": "yupup", "extra": "yup"})

    with pytest.raises(AssertionError):
        _validate(_single_required_string_config_dict(), {"string_field": 1})


def test_undefined_field_error():
    with pytest.raises(
        AssertionError,
        match=(
            'Received unexpected config entry "extra" at the root. Expected: "{ string_field: '
            'String }".'
        ),
    ):
        _validate(
            _single_required_string_config_dict(), {"string_field": "value", "extra": "extra"}
        )


def test_multiple_required_fields_passing():
    assert (
        _validate(
            _multiple_required_fields_config_dict(),
            {"field_one": "value_one", "field_two": "value_two"},
        )
        == {"field_one": "value_one", "field_two": "value_two"}
    )


def test_multiple_required_fields_failing():
    with pytest.raises(AssertionError):
        _validate(_multiple_required_fields_config_dict(), {})

    with pytest.raises(AssertionError):
        _validate(_multiple_required_fields_config_dict(), {"field_one": "yup"})

    with pytest.raises(AssertionError):
        _validate(_multiple_required_fields_config_dict(), {"field_one": "yup", "extra": "yup"})

    with pytest.raises(AssertionError):
        _validate(
            _multiple_required_fields_config_dict(),
            {"field_one": "yup", "field_two": "yup", "extra": "should_not_exist"},
        )

    with pytest.raises(AssertionError):
        _validate(
            _multiple_required_fields_config_dict(), {"field_one": "value_one", "field_two": 2}
        )


def test_single_optional_field_passing():
    assert _validate(_single_optional_string_config_dict(), {"optional_field": "value"}) == {
        "optional_field": "value"
    }
    assert _validate(_single_optional_string_config_dict(), {}) == {}

    with pytest.raises(AssertionError):
        assert _validate(_single_optional_string_config_dict(), {"optional_field": None}) == {
            "optional_field": None
        }


def test_single_optional_field_failing():
    with pytest.raises(AssertionError):

        _validate(_single_optional_string_config_dict(), {"optional_field": 1})

    with pytest.raises(AssertionError):
        _validate(_single_optional_string_config_dict(), {"dlkjfalksdjflksaj": 1})


def test_single_optional_field_passing_with_default():
    assert _validate(_single_optional_string_field_config_dict_with_default(), {}) == {
        "optional_field": "some_default"
    }

    assert _validate(
        _single_optional_string_field_config_dict_with_default(), {"optional_field": "override"}
    ) == {"optional_field": "override"}


def test_permissive_multiple_required_fields_passing():
    assert _validate(
        _multiple_required_fields_config_permissive_dict(),
        {
            "field_one": "value_one",
            "field_two": "value_two",
            "previously_unspecified": "should_exist",
        },
    ) == {
        "field_one": "value_one",
        "field_two": "value_two",
        "previously_unspecified": "should_exist",
    }


def test_permissive_multiple_required_fields_nested_passing():
    assert _validate(
        _multiple_required_fields_config_permissive_dict(),
        {
            "field_one": "value_one",
            "field_two": "value_two",
            "previously_unspecified": {"nested": "value", "with_int": 2},
        },
    ) == {
        "field_one": "value_one",
        "field_two": "value_two",
        "previously_unspecified": {"nested": "value", "with_int": 2},
    }


def test_permissive_multiple_required_fields_failing():
    with pytest.raises(AssertionError):
        _validate(_multiple_required_fields_config_permissive_dict(), {})

    with pytest.raises(AssertionError):
        _validate(_multiple_required_fields_config_permissive_dict(), {"field_one": "yup"})

    with pytest.raises(AssertionError):
        _validate(
            _multiple_required_fields_config_permissive_dict(),
            {"field_one": "value_one", "field_two": 2},
        )


def test_mixed_args_passing():
    assert (
        _validate(
            _mixed_required_optional_string_config_dict_with_default(),
            {"optional_arg": "value_one", "required_arg": "value_two"},
        )
        == {"optional_arg": "value_one", "required_arg": "value_two"}
    )

    assert _validate(
        _mixed_required_optional_string_config_dict_with_default(), {"required_arg": "value_two"}
    ) == {"optional_arg": "some_default", "required_arg": "value_two"}

    assert _validate(
        _mixed_required_optional_string_config_dict_with_default(),
        {"required_arg": "value_two", "optional_arg_no_default": "value_three"},
    ) == {
        "optional_arg": "some_default",
        "required_arg": "value_two",
        "optional_arg_no_default": "value_three",
    }


def _single_nested_config():
    return convert_potential_field({"nested": {"int_field": Int}})


def _nested_optional_config_with_default():
    return convert_potential_field(
        {"nested": {"int_field": Field(Int, is_required=False, default_value=3)}}
    )


def _nested_optional_config_with_no_default():
    return convert_potential_field({"nested": {"int_field": Field(Int, is_required=False)}})


def test_single_nested_config():
    assert _validate(_single_nested_config(), {"nested": {"int_field": 2}}) == {
        "nested": {"int_field": 2}
    }


def test_single_nested_config_undefined_errors():
    with pytest.raises(
        AssertionError,
        match='Value at path root:nested must be dict. Expected: "{ int_field: Int }".',
    ):
        _validate(_single_nested_config(), {"nested": "dkjfdk"})

    with pytest.raises(
        AssertionError,
        match='Invalid scalar at path root:nested:int_field. Value "dkjfdk" of type .* is not valid for expected type "Int".',
    ):
        _validate(_single_nested_config(), {"nested": {"int_field": "dkjfdk"}})

    with pytest.raises(
        AssertionError,
        match=(
            'Received unexpected config entry "not_a_field" at path root:nested. Expected: '
            '"{ int_field: Int }".'
        ),
    ):
        _validate(_single_nested_config(), {"nested": {"int_field": 2, "not_a_field": 1}})

    with pytest.raises(
        AssertionError,
        match="Invalid scalar at path root:nested:int_field. Value \"{'too_nested': 'dkjfdk'}\" of type .* is not valid for expected type \"Int\".",
    ):
        _validate(_single_nested_config(), {"nested": {"int_field": {"too_nested": "dkjfdk"}}})


def test_nested_optional_with_default():
    assert _validate(_nested_optional_config_with_default(), {"nested": {"int_field": 2}}) == {
        "nested": {"int_field": 2}
    }

    assert _validate(_nested_optional_config_with_default(), {"nested": {}}) == {
        "nested": {"int_field": 3}
    }


def test_nested_optional_with_no_default():
    assert _validate(_nested_optional_config_with_no_default(), {"nested": {"int_field": 2}}) == {
        "nested": {"int_field": 2}
    }

    assert _validate(_nested_optional_config_with_no_default(), {"nested": {}}) == {"nested": {}}


def test_config_defaults():
    @solid(config_schema={"sum": Int})
    def two(_context):
        assert _context.solid_config["sum"] == 6
        return _context.solid_config["sum"]

    @solid(config_schema={"sum": Int})
    def one(_context, prev_sum):
        assert prev_sum == 6
        return prev_sum + _context.solid_config["sum"]

    # addition_composite_solid
    def addition_composite_solid_config_fn(config):
        child_config = {"config": {"sum": config["a"] + config["b"] + config["c"]}}
        return {"one": child_config, "two": child_config}

    @composite_solid(
        config_fn=addition_composite_solid_config_fn,
        config_schema={
            "a": Field(Int, is_required=False, default_value=1),
            "b": Field(Int, is_required=False, default_value=2),
            "c": Int,
        },
    )
    def addition_composite_solid():
        return one(two())

    @pipeline
    def addition_pipeline():
        addition_composite_solid()

    result = execute_pipeline(
        addition_pipeline, {"solids": {"addition_composite_solid": {"config": {"c": 3}}}}
    )

    assert result.success


def test_config_with_and_without_config():
    @solid(config_schema={"prefix": Field(str, is_required=False, default_value="_")})
    def prefix_value(context, v):
        return "{prefix}{v}".format(prefix=context.solid_config["prefix"], v=v)

    @composite_solid(
        config_fn=lambda cfg: {"prefix_value": {"config": {"prefix": cfg["prefix"]}}},
        config_schema={"prefix": Field(str, is_required=False, default_value="_id_")},
    )
    def prefix_id(val):
        return prefix_value(val)

    @solid
    def print_value(_, v):
        return str(v)

    @pipeline
    def config_issue_pipeline():
        v = prefix_id()
        print_value(v)

    result = execute_pipeline(
        config_issue_pipeline,
        {
            "solids": {
                "prefix_id": {
                    "config": {"prefix": "_customprefix_"},
                    "inputs": {"val": {"value": "12345"}},
                }
            }
        },
    )

    assert result.success
    assert result.result_for_solid("print_value").output_value() == "_customprefix_12345"

    result_using_default = execute_pipeline(
        config_issue_pipeline,
        {"solids": {"prefix_id": {"config": {}, "inputs": {"val": {"value": "12345"}}}}},
    )

    assert result_using_default.success
    assert result_using_default.result_for_solid("print_value").output_value() == "_id_12345"


def test_build_optionality():
    optional_test_type = convert_potential_field(
        {
            "required": {"value": String},
            "optional": {"value": Field(String, is_required=False)},
        }
    ).config_type

    assert optional_test_type.fields["required"].is_required
    assert optional_test_type.fields["optional"].is_required is False


def test_wrong_solid_name():
    @solid(name="some_solid", input_defs=[], output_defs=[], config_schema=Int)
    def some_solid(_):
        return None

    @pipeline(name="pipeline_wrong_solid_name")
    def pipeline_def():
        some_solid()

    env_config = {"solids": {"another_name": {"config": {}}}}

    with pytest.raises(DagsterInvalidConfigError) as pe_info:
        execute_pipeline(pipeline_def, env_config)

    pe = pe_info.value

    assert 'Received unexpected config entry "another_name" at path root:solids' in str(pe)


def fail_me():
    assert False


def dummy_resource(config_schema=None):
    return ResourceDefinition(lambda _: None, config_schema=config_schema)


def test_wrong_resources():
    pipeline_def = PipelineDefinition(
        name="pipeline_test_multiple_context",
        mode_defs=[
            ModeDefinition(
                resource_defs={"resource_one": dummy_resource(), "resource_two": dummy_resource()}
            )
        ],
        solid_defs=[],
    )

    with pytest.raises(
        DagsterInvalidConfigError,
        match='Received unexpected config entry "nope" at path root:resources',
    ):
        execute_pipeline(pipeline_def, {"resources": {"nope": {}}})


def test_solid_list_config():
    value = [1, 2]
    called = {}

    @solid(name="solid_list_config", input_defs=[], output_defs=[], config_schema=[int])
    def solid_list_config(context):
        assert context.solid_config == value
        called["yup"] = True

    @pipeline(name="solid_list_config_pipeline")
    def pipeline_def():
        solid_list_config()

    result = execute_pipeline(
        pipeline_def, run_config={"solids": {"solid_list_config": {"config": value}}}
    )

    assert result.success
    assert called["yup"]


def test_two_list_types():
    @solid(
        input_defs=[],
        config_schema={"list_one": [int], "list_two": [int]},
    )
    def two_list_type(context):
        return context.solid_config

    assert (
        execute_solid(
            two_list_type,
            run_config={
                "solids": {"two_list_type": {"config": {"list_one": [1], "list_two": [2]}}}
            },
        ).output_value()
        == {"list_one": [1], "list_two": [2]}
    )

    @solid(
        input_defs=[],
        config_schema={"list_one": [Int], "list_two": [Int]},
    )
    def two_list_type_condensed_syntax(context):
        return context.solid_config

    assert (
        execute_solid(
            two_list_type_condensed_syntax,
            run_config={
                "solids": {
                    "two_list_type_condensed_syntax": {"config": {"list_one": [1], "list_two": [2]}}
                }
            },
        ).output_value()
        == {"list_one": [1], "list_two": [2]}
    )

    @solid(
        input_defs=[],
        config_schema={"list_one": [int], "list_two": [int]},
    )
    def two_list_type_condensed_syntax_primitives(context):
        return context.solid_config

    assert (
        execute_solid(
            two_list_type_condensed_syntax_primitives,
            run_config={
                "solids": {
                    "two_list_type_condensed_syntax_primitives": {
                        "config": {"list_one": [1], "list_two": [2]}
                    }
                }
            },
        ).output_value()
        == {"list_one": [1], "list_two": [2]}
    )


def test_multilevel_default_handling():
    @solid(config_schema=Field(Int, is_required=False, default_value=234))
    def has_default_value(context):
        assert context.solid_config == 234

    pipeline_def = PipelineDefinition(
        name="multilevel_default_handling", solid_defs=[has_default_value]
    )

    assert execute_pipeline(pipeline_def).success
    assert execute_pipeline(pipeline_def, run_config=None).success
    assert execute_pipeline(pipeline_def, run_config={}).success
    assert execute_pipeline(pipeline_def, run_config={"solids": {}}).success
    assert execute_pipeline(pipeline_def, run_config={"solids": {"has_default_value": {}}}).success

    assert execute_pipeline(
        pipeline_def, run_config={"solids": {"has_default_value": {"config": 234}}}
    ).success


def test_no_env_missing_required_error_handling():
    @solid(config_schema=Int)
    def required_int_solid(_context):
        pass

    pipeline_def = PipelineDefinition(
        name="no_env_missing_required_error", solid_defs=[required_int_solid]
    )

    with pytest.raises(DagsterInvalidConfigError) as pe_info:
        execute_pipeline(pipeline_def)

    assert isinstance(pe_info.value, DagsterInvalidConfigError)
    pe = pe_info.value
    assert len(pe.errors) == 1
    mfe = pe.errors[0]
    assert mfe.reason == DagsterEvaluationErrorReason.MISSING_REQUIRED_FIELD
    assert len(pe.errors) == 1

    expected_suggested_config = {"solids": {"required_int_solid": {"config": 0}}}
    assert pe.errors[0].message.startswith('Missing required config entry "solids" at the root.')
    assert str(expected_suggested_config) in pe.errors[0].message


def test_root_extra_field():
    @solid(config_schema=Int)
    def required_int_solid(_context):
        pass

    @pipeline
    def pipeline_def():
        required_int_solid()

    with pytest.raises(DagsterInvalidConfigError) as pe_info:
        execute_pipeline(
            pipeline_def,
            run_config={"solids": {"required_int_solid": {"config": 948594}}, "nope": None},
        )

    pe = pe_info.value
    assert len(pe.errors) == 1
    fnd = pe.errors[0]
    assert fnd.reason == DagsterEvaluationErrorReason.FIELD_NOT_DEFINED
    assert 'Received unexpected config entry "nope"' in pe.message


def test_deeper_path():
    @solid(config_schema=Int)
    def required_int_solid(_context):
        pass

    @pipeline
    def pipeline_def():
        required_int_solid()

    with pytest.raises(DagsterInvalidConfigError) as pe_info:
        execute_pipeline(
            pipeline_def, run_config={"solids": {"required_int_solid": {"config": "asdf"}}}
        )

    pe = pe_info.value
    assert len(pe.errors) == 1
    rtm = pe.errors[0]
    assert rtm.reason == DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH


def test_working_list_path():
    called = {}

    @solid(config_schema=[int])
    def required_list_int_solid(context):
        assert context.solid_config == [1, 2]
        called["yup"] = True

    @pipeline
    def pipeline_def():
        required_list_int_solid()

    result = execute_pipeline(
        pipeline_def, run_config={"solids": {"required_list_int_solid": {"config": [1, 2]}}}
    )

    assert result.success
    assert called["yup"]


def test_item_error_list_path():
    called = {}

    @solid(config_schema=[int])
    def required_list_int_solid(context):
        assert context.solid_config == [1, 2]
        called["yup"] = True

    @pipeline
    def pipeline_def():
        required_list_int_solid()

    with pytest.raises(DagsterInvalidConfigError) as pe_info:
        execute_pipeline(
            pipeline_def,
            run_config={"solids": {"required_list_int_solid": {"config": [1, "nope"]}}},
        )

    pe = pe_info.value
    assert len(pe.errors) == 1
    rtm = pe.errors[0]
    assert rtm.reason == DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH

    assert "Invalid scalar at path root:solids:required_list_int_solid:config[1]" in str(pe)


def test_list_in_config_error():
    error_msg = (
        "Cannot use List in the context of config. "
        "Please use a python list (e.g. [int]) or dagster.Array (e.g. Array(int)) instead."
    )

    with pytest.raises(DagsterInvalidDefinitionError, match=re.escape(error_msg)):

        @solid(config_schema=List[int])
        def _no_runtime_list_in_config(_):
            pass


def test_required_resource_not_given():
    @solid(required_resource_keys={"required"})
    def needs_resource(_):
        pass

    @pipeline(
        name="required_resource_not_given",
        mode_defs=[ModeDefinition(resource_defs={"required": dummy_resource(Int)})],
    )
    def pipeline_def():
        needs_resource()

    with pytest.raises(DagsterInvalidConfigError) as not_none_pe_info:
        execute_pipeline(pipeline_def, run_config={"resources": None})

    assert len(not_none_pe_info.value.errors) == 1
    assert (
        "Value at path root:resources must not be None." in not_none_pe_info.value.errors[0].message
    )

    with pytest.raises(DagsterInvalidConfigError) as pe_info:
        execute_pipeline(pipeline_def, run_config={"resources": {}})

    pe = pe_info.value
    error = pe.errors[0]
    assert error.reason == DagsterEvaluationErrorReason.MISSING_REQUIRED_FIELD

    expected_suggested_config = {"required": {"config": 0}}
    assert error.message.startswith(
        'Missing required config entry "required" at path root:resources.'
    )
    assert str(expected_suggested_config) in error.message


def test_multilevel_good_error_handling_solids():
    @solid(config_schema=Int)
    def good_error_handling(_context):
        pass

    @pipeline
    def pipeline_def():
        good_error_handling()

    with pytest.raises(DagsterInvalidConfigError) as not_none_pe_info:
        execute_pipeline(pipeline_def, run_config={"solids": None})

    assert len(not_none_pe_info.value.errors) == 1
    assert "Value at path root:solids must not be None." in not_none_pe_info.value.errors[0].message

    with pytest.raises(DagsterInvalidConfigError) as missing_field_pe_info:
        execute_pipeline(pipeline_def, run_config={"solids": {}})

    assert len(missing_field_pe_info.value.errors) == 1

    expected_suggested_config = {"good_error_handling": {"config": 0}}
    assert missing_field_pe_info.value.errors[0].message.startswith(
        """Missing required config entry "good_error_handling" at path root:solids."""
    )
    assert str(expected_suggested_config) in missing_field_pe_info.value.errors[0].message


def test_multilevel_good_error_handling_solid_name_solids():
    @solid(config_schema=Int)
    def good_error_handling(_context):
        pass

    @pipeline
    def pipeline_def():
        good_error_handling()

    with pytest.raises(DagsterInvalidConfigError) as pe_info:
        execute_pipeline(pipeline_def, run_config={"solids": {"good_error_handling": {}}})

    assert len(pe_info.value.errors) == 1

    expected_suggested_config = {"config": 0}
    assert pe_info.value.errors[0].message.startswith(
        """Missing required config entry "config" at path root:solids:good_error_handling."""
    )
    assert str(expected_suggested_config) in pe_info.value.errors[0].message


def test_multilevel_good_error_handling_config_solids_name_solids():
    @solid(config_schema=Noneable(int))
    def good_error_handling(_context):
        pass

    @pipeline
    def pipeline_def():
        good_error_handling()

    execute_pipeline(pipeline_def, run_config={"solids": {"good_error_handling": {"config": None}}})


def test_invalid_default_values():
    with pytest.raises(
        DagsterInvalidConfigError,
        match='Value "3" of type .* is not valid for expected type "Int"',
    ):

        @solid(config_schema=Field(Int, default_value="3"))
        def _solid(_):
            pass


def test_typing_types_into_config():
    match_str = re.escape(
        "You have passed in typing.List to the config system. "
        "Types from the typing module in python are not allowed "
        "in the config system. You must use types that are imported "
        "from dagster or primitive types such as bool, int, etc."
    )
    with pytest.raises(DagsterInvalidDefinitionError, match=match_str):

        @solid(config_schema=Field(typing.List))
        def _solid(_):
            pass

    with pytest.raises(DagsterInvalidDefinitionError, match=match_str):

        @solid(config_schema=typing.List)
        def _solid(_):
            pass

    match_str = re.escape(
        "You have passed in typing.List[int] to the config system. Types "
        "from the typing module in python are not allowed in the config system. "
        "You must use types that are imported from dagster or primitive types "
        "such as bool, int, etc."
    )

    with pytest.raises(DagsterInvalidDefinitionError, match=match_str):

        @solid(config_schema=Field(typing.List[int]))
        def _solid(_):
            pass

    with pytest.raises(DagsterInvalidDefinitionError, match=match_str):

        @solid(config_schema=typing.List[int])
        def _solid(_):
            pass

    for ttype in [
        typing.Optional[int],
        typing.Set,
        typing.Set[int],
        typing.Dict,
        typing.Dict[int, str],
        typing.Tuple,
        typing.Tuple[int, int],
    ]:
        with pytest.raises(DagsterInvalidDefinitionError):

            @solid(config_schema=Field(ttype))
            def _solid(_):
                pass


def test_no_set_in_config_system():
    set_error_msg = re.escape("Cannot use Set in the context of a config field.")
    with pytest.raises(DagsterInvalidDefinitionError, match=set_error_msg):

        @solid(config_schema=Field(Set))
        def _bare_open_set(_):
            pass

    with pytest.raises(DagsterInvalidDefinitionError, match=set_error_msg):

        @solid(config_schema=Set)
        def _bare_open_set(_):
            pass

    with pytest.raises(DagsterInvalidDefinitionError, match=set_error_msg):

        @solid(config_schema=Field(Set[int]))
        def _bare_closed_set(_):
            pass

    with pytest.raises(DagsterInvalidDefinitionError, match=set_error_msg):

        @solid(config_schema=Set[int])
        def _bare_closed_set(_):
            pass


def test_no_tuple_in_config_system():
    tuple_error_msg = re.escape("Cannot use Tuple in the context of a config field.")
    with pytest.raises(DagsterInvalidDefinitionError, match=tuple_error_msg):

        @solid(config_schema=Field(Tuple))
        def _bare_open_tuple(_):
            pass

    with pytest.raises(DagsterInvalidDefinitionError, match=tuple_error_msg):

        @solid(config_schema=Field(Tuple[int]))
        def _bare_closed_set(_):
            pass


def test_field_is_none():
    with pytest.raises(DagsterInvalidConfigDefinitionError) as exc_info:

        @solid(config_schema={"none_field": None})
        def _none_is_bad(_):
            pass

    assert "Fields cannot be None" in str(exc_info.value)


def test_permissive_defaults():
    @solid(config_schema=Permissive({"four": Field(int, default_value=4)}))
    def perm_with_defaults(context):
        assert context.solid_config["four"] == 4

    assert execute_solid(perm_with_defaults).success


def test_permissive_ordering():
    alphabet = {letter: letter for letter in string.ascii_lowercase}

    @solid(config_schema=dict)
    def test_order(context):
        alpha = list(context.solid_config.keys())
        for idx, letter in enumerate(string.ascii_lowercase):
            assert letter == alpha[idx]  # if this fails dict ordering got messed up

    assert execute_solid(
        test_order, run_config={"solids": {"test_order": {"config": alphabet}}}
    ).success



import pytest
from dagster import Field, resource
from dagster.core.errors import DagsterResourceFunctionError
from dagster.core.execution.build_resources import build_resources
from dagster.core.storage.mem_io_manager import InMemoryIOManager


def test_basic_resource():
    @resource
    def basic_resource(_):
        return "foo"

    with build_resources(
        resources={"basic_resource": basic_resource},
    ) as resources:
        assert resources.basic_resource == "foo"


def test_resource_with_config():
    @resource(
        config_schema={"plant": str, "animal": Field(str, is_required=False, default_value="dog")}
    )
    def basic_resource(init_context):
        plant = init_context.resource_config["plant"]
        animal = init_context.resource_config["animal"]
        return f"plant: {plant}, animal: {animal}"

    with build_resources(
        resources={"basic_resource": basic_resource},
        run_config={"basic_resource": {"config": {"plant": "maple tree"}}},
    ) as resources:
        assert resources.basic_resource == "plant: maple tree, animal: dog"


def test_resource_with_dependencies():
    @resource(config_schema={"animal": str})
    def no_deps(init_context):
        return init_context.resource_config["animal"]

    @resource(required_resource_keys={"no_deps"})
    def has_deps(init_context):
        return f"{init_context.resources.no_deps} is an animal."

    with build_resources(
        resources={"no_deps": no_deps, "has_deps": has_deps},
        run_config={"no_deps": {"config": {"animal": "dog"}}},
    ) as resources:
        assert resources.no_deps == "dog"
        assert resources.has_deps == "dog is an animal."


def test_error_in_resource_initialization():
    @resource
    def i_will_fail(_):
        raise Exception("Failed.")

    with pytest.raises(
        DagsterResourceFunctionError,
        match="Error executing resource_fn on ResourceDefinition i_will_fail",
    ):
        with build_resources(resources={"i_will_fail": i_will_fail}):
            pass


# Ensure that build_resources can provide an instance to the initialization process of a resource.
def test_resource_init_requires_instance():
    @resource
    def basic_resource(init_context):
        assert init_context.instance

    build_resources({"basic_resource": basic_resource})


def test_resource_init_values():
    @resource(required_resource_keys={"bar"})
    def foo_resource(init_context):
        assert init_context.resources.bar == "bar"
        return "foo"

    with build_resources({"foo": foo_resource, "bar": "bar"}) as resources:
        assert resources.foo == "foo"
        assert resources.bar == "bar"


def test_resource_init_io_manager():
    build_resources({"io_manager": InMemoryIOManager})



import json
import logging
import re
from contextlib import contextmanager

import pytest
from dagster import ModeDefinition, check, execute_solid, pipeline, resource, solid
from dagster.core.definitions import SolidHandle
from dagster.core.events import DagsterEvent
from dagster.core.execution.context.logger import InitLoggerContext
from dagster.core.execution.plan.objects import StepFailureData
from dagster.core.execution.plan.outputs import StepOutputHandle
from dagster.core.log_manager import DagsterLogManager
from dagster.loggers import colored_console_logger, json_console_logger
from dagster.utils.error import SerializableErrorInfo

REGEX_UUID = r"[a-z-0-9]{8}\-[a-z-0-9]{4}\-[a-z-0-9]{4}\-[a-z-0-9]{4}\-[a-z-0-9]{12}"
REGEX_TS = r"\d{4}\-\d{2}\-\d{2}T\d{2}:\d{2}:\d{2}\.\d{6}"

DAGSTER_DEFAULT_LOGGER = "dagster"


@contextmanager
def _setup_logger(name, log_levels=None):
    """Test helper that creates a new logger.

    Args:
        name (str): The name of the logger.
        log_levels (Optional[Dict[str, int]]): Any non-standard log levels to expose on the logger
            (e.g., logger.success)
    """
    log_levels = check.opt_dict_param(log_levels, "log_levels")

    class TestLogger(logging.Logger):  # py27 compat
        pass

    logger = TestLogger(name)

    captured_results = []

    def log_fn(msg, *args, **kwargs):  # pylint:disable=unused-argument
        captured_results.append(msg)

    def int_log_fn(lvl, msg, *args, **kwargs):  # pylint:disable=unused-argument
        captured_results.append(msg)

    for level in ["debug", "info", "warning", "error", "critical"] + list(
        [x.lower() for x in log_levels.keys()]
    ):
        setattr(logger, level, log_fn)
        setattr(logger, "log", int_log_fn)

    yield (captured_results, logger)


def _regex_match_kv_pair(regex, kv_pairs):
    return any([re.match(regex, kv_pair) for kv_pair in kv_pairs])


def test_logging_no_loggers_registered():
    dl = DagsterLogManager("none", {}, [])
    dl.debug("test")
    dl.info("test")
    dl.warning("test")
    dl.error("test")
    dl.critical("test")


def test_logging_basic():
    with _setup_logger("test") as (captured_results, logger):

        dl = DagsterLogManager("123", {}, [logger])
        dl.debug("test")
        dl.info("test")
        dl.warning("test")
        dl.error("test")
        dl.critical("test")

        assert captured_results == ["system - 123 - test"] * 5


def test_logging_custom_log_levels():
    with _setup_logger("test", {"FOO": 3}) as (_captured_results, logger):

        dl = DagsterLogManager("123", {}, [logger])
        with pytest.raises(AttributeError):
            dl.foo("test")  # pylint: disable=no-member


def test_logging_integer_log_levels():
    with _setup_logger("test", {"FOO": 3}) as (_captured_results, logger):

        dl = DagsterLogManager("123", {}, [logger])
        dl.log(3, "test")  # pylint: disable=no-member


def test_logging_bad_custom_log_levels():
    with _setup_logger("test") as (_, logger):

        dl = DagsterLogManager("123", {}, [logger])
        with pytest.raises(check.CheckError):
            dl._log("test", "foobar", {})  # pylint: disable=protected-access


def test_multiline_logging_complex():
    msg = "DagsterEventType.STEP_FAILURE for step start.materialization.output.result.0"
    kwargs = {
        "pipeline": "example",
        "pipeline_name": "error_monster",
        "step_key": "start.materialization.output.result.0",
        "solid": "start",
        "solid_definition": "emit_num",
        "dagster_event": DagsterEvent(
            event_type_value="STEP_FAILURE",
            pipeline_name="error_monster",
            step_key="start.materialization.output.result.0",
            solid_handle=SolidHandle("start", None),
            step_kind_value="MATERIALIZATION_THUNK",
            logging_tags={
                "pipeline": "error_monster",
                "step_key": "start.materialization.output.result.0",
                "solid": "start",
                "solid_definition": "emit_num",
            },
            event_specific_data=StepFailureData(
                error=SerializableErrorInfo(
                    message="FileNotFoundError: [Errno 2] No such file or directory: '/path/to/file'\n",
                    stack=["a stack message"],
                    cls_name="FileNotFoundError",
                ),
                user_failure_data=None,
            ),
        ),
    }

    with _setup_logger(DAGSTER_DEFAULT_LOGGER) as (captured_results, logger):

        dl = DagsterLogManager("123", {}, [logger])
        dl.info(msg, **kwargs)

    expected_results = [
        "error_monster - 123 - STEP_FAILURE - DagsterEventType.STEP_FAILURE for step "
        "start.materialization.output.result.0",
        "",
        "FileNotFoundError: [Errno 2] No such file or directory: '/path/to/file'",
        "",
        "Stack Trace:",
        "a stack message",
    ]

    assert captured_results[0].split("\n") == expected_results


def test_default_context_logging():
    called = {}

    @solid(input_defs=[], output_defs=[])
    def default_context_solid(context):
        called["yes"] = True
        for logger in context.log.loggers:
            assert logger.level == logging.DEBUG

    execute_solid(default_context_solid)

    assert called["yes"]


def test_colored_console_logger_with_integer_log_level():
    @pipeline
    def pipe():
        pass

    colored_console_logger.logger_fn(
        InitLoggerContext(
            {"name": "dagster", "log_level": 4},
            colored_console_logger,
            pipeline_def=pipe,
        )
    )


def test_json_console_logger(capsys):
    @solid
    def hello_world(context):
        context.log.info("Hello, world!")

    execute_solid(
        hello_world,
        mode_def=ModeDefinition(logger_defs={"json": json_console_logger}),
        run_config={"loggers": {"json": {"config": {}}}},
    )

    captured = capsys.readouterr()

    found_msg = False
    for line in captured.err.split("\n"):
        if line:
            parsed = json.loads(line)
            if parsed["dagster_meta"]["orig_message"] == "Hello, world!":
                found_msg = True

    assert found_msg


def test_resource_logging(capsys):
    @resource
    def foo_resource(init_context):
        def fn():
            init_context.log_manager.info("test logging from foo resource")

        return fn

    @resource
    def bar_resource(init_context):
        def fn():
            init_context.log_manager.info("test logging from bar resource")

        return fn

    @solid(required_resource_keys={"foo", "bar"})
    def process(context):
        context.resources.foo()
        context.resources.bar()

    execute_solid(
        process,
        mode_def=ModeDefinition(resource_defs={"foo": foo_resource, "bar": bar_resource}),
    )

    captured = capsys.readouterr()

    expected_log_regexes = [
        r"dagster - INFO - resource:foo - [a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-"
        r"[a-f0-9]{12} - process - test logging from foo resource",
        r"dagster - INFO - resource:bar - [a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-"
        r"[a-f0-9]{12} - process - test logging from bar resource",
    ]
    for expected_log_regex in expected_log_regexes:
        assert re.search(expected_log_regex, captured.err, re.MULTILINE)


def test_io_context_logging(capsys):
    @solid
    def logged_solid(context):
        context.get_step_execution_context().get_output_context(
            StepOutputHandle("logged_solid", "result")
        ).log.debug("test OUTPUT debug logging from logged_solid.")
        context.get_step_execution_context().for_input_manager(
            "logged_solid", {}, {}, None, source_handle=None
        ).log.debug("test INPUT debug logging from logged_solid.")

    result = execute_solid(logged_solid)
    assert result.success

    captured = capsys.readouterr()

    assert re.search("test OUTPUT debug logging from logged_solid.", captured.err, re.MULTILINE)
    assert re.search("test INPUT debug logging from logged_solid.", captured.err, re.MULTILINE)



import pytest
from dagster import (
    Any,
    DagsterInvalidDefinitionError,
    DependencyDefinition,
    InputDefinition,
    Int,
    List,
    MultiDependencyDefinition,
    Nothing,
    OutputDefinition,
    PipelineDefinition,
    composite_solid,
    execute_pipeline,
    lambda_solid,
    pipeline,
    solid,
)
from dagster.core.definitions.composition import MappedInputPlaceholder
from dagster.core.definitions.solid import CompositeSolidDefinition


def test_simple_values():
    @solid(input_defs=[InputDefinition("numbers", List[Int])])
    def sum_num(_context, numbers):
        # cant guarantee order
        assert set(numbers) == set([1, 2, 3])
        return sum(numbers)

    @lambda_solid
    def emit_1():
        return 1

    @lambda_solid
    def emit_2():
        return 2

    @lambda_solid
    def emit_3():
        return 3

    result = execute_pipeline(
        PipelineDefinition(
            name="input_test",
            solid_defs=[emit_1, emit_2, emit_3, sum_num],
            dependencies={
                "sum_num": {
                    "numbers": MultiDependencyDefinition(
                        [
                            DependencyDefinition("emit_1"),
                            DependencyDefinition("emit_2"),
                            DependencyDefinition("emit_3"),
                        ]
                    )
                }
            },
        )
    )
    assert result.success
    assert result.result_for_solid("sum_num").output_value() == 6


@solid(input_defs=[InputDefinition("stuff", List[Any])])
def collect(_context, stuff):
    assert set(stuff) == set([1, None, "one"])
    return stuff


@lambda_solid
def emit_num():
    return 1


@lambda_solid
def emit_none():
    pass


@lambda_solid
def emit_str():
    return "one"


@lambda_solid(output_def=OutputDefinition(Nothing))
def emit_nothing():
    pass


def test_interleaved_values():
    result = execute_pipeline(
        PipelineDefinition(
            name="input_test",
            solid_defs=[emit_num, emit_none, emit_str, collect],
            dependencies={
                "collect": {
                    "stuff": MultiDependencyDefinition(
                        [
                            DependencyDefinition("emit_num"),
                            DependencyDefinition("emit_none"),
                            DependencyDefinition("emit_str"),
                        ]
                    )
                }
            },
        )
    )
    assert result.success


def test_dsl():
    @pipeline
    def input_test():
        collect([emit_num(), emit_none(), emit_str()])

    result = execute_pipeline(input_test)

    assert result.success


def test_collect_one():
    @lambda_solid
    def collect_one(list_arg):
        assert list_arg == ["one"]

    @pipeline
    def multi_one():
        collect_one([emit_str()])

    assert execute_pipeline(multi_one).success


def test_fan_in_manual():
    # manually building up this guy
    @composite_solid
    def _target_composite_dsl(str_in, none_in):
        num = emit_num()
        return collect([num, str_in, none_in])

    # base case works
    _target_composite_manual = CompositeSolidDefinition(
        name="manual_composite",
        solid_defs=[emit_num, collect],
        input_mappings=[
            InputDefinition("str_in").mapping_to("collect", "stuff", 1),
            InputDefinition("none_in").mapping_to("collect", "stuff", 2),
        ],
        output_mappings=[OutputDefinition().mapping_from("collect")],
        dependencies={
            "collect": {
                "stuff": MultiDependencyDefinition(
                    [
                        DependencyDefinition("emit_num"),
                        MappedInputPlaceholder,
                        MappedInputPlaceholder,
                    ]
                )
            }
        },
    )

    with pytest.raises(
        DagsterInvalidDefinitionError,
        match="index 2 in the MultiDependencyDefinition is not a MappedInputPlaceholder",
    ):
        _missing_placeholder = CompositeSolidDefinition(
            name="manual_composite",
            solid_defs=[emit_num, collect],
            input_mappings=[
                InputDefinition("str_in").mapping_to("collect", "stuff", 1),
                InputDefinition("none_in").mapping_to("collect", "stuff", 2),
            ],
            output_mappings=[OutputDefinition().mapping_from("collect")],
            dependencies={
                "collect": {
                    "stuff": MultiDependencyDefinition(
                        [
                            DependencyDefinition("emit_num"),
                            MappedInputPlaceholder,
                        ]
                    )
                }
            },
        )

    with pytest.raises(DagsterInvalidDefinitionError, match="is not a MultiDependencyDefinition"):
        _bad_target = CompositeSolidDefinition(
            name="manual_composite",
            solid_defs=[emit_num, collect],
            input_mappings=[
                InputDefinition("str_in").mapping_to("collect", "stuff", 1),
                InputDefinition("none_in").mapping_to("collect", "stuff", 2),
            ],
            output_mappings=[OutputDefinition().mapping_from("collect")],
            dependencies={"collect": {"stuff": DependencyDefinition("emit_num")}},
        )

    with pytest.raises(
        DagsterInvalidDefinitionError,
        match="Unsatisfied MappedInputPlaceholder at index 3",
    ):
        _missing_placeholder = CompositeSolidDefinition(
            name="manual_composite",
            solid_defs=[emit_num, collect],
            input_mappings=[
                InputDefinition("str_in").mapping_to("collect", "stuff", 1),
                InputDefinition("none_in").mapping_to("collect", "stuff", 2),
            ],
            output_mappings=[OutputDefinition().mapping_from("collect")],
            dependencies={
                "collect": {
                    "stuff": MultiDependencyDefinition(
                        [
                            DependencyDefinition("emit_num"),
                            MappedInputPlaceholder,
                            MappedInputPlaceholder,
                            MappedInputPlaceholder,
                        ]
                    )
                }
            },
        )


def test_nothing_deps():

    with pytest.raises(
        DagsterInvalidDefinitionError,
        match=r'Input "stuff" expects a value of type \[Any\] and output '
        '"result" returns type Nothing',
    ):
        PipelineDefinition(
            name="input_test",
            solid_defs=[emit_num, emit_nothing, emit_str, collect],
            dependencies={
                "collect": {
                    "stuff": MultiDependencyDefinition(
                        [
                            DependencyDefinition("emit_num"),
                            DependencyDefinition("emit_nothing"),
                            DependencyDefinition("emit_str"),
                        ]
                    )
                }
            },
        )



import re

import pytest
from dagster import DagsterInvalidDefinitionError, solid
from dagster.core.definitions.decorators.graph import graph
from dagster.core.execution.execute import execute_in_process
from dagster.experimental import DynamicOutput, DynamicOutputDefinition


def get_solids():
    @solid
    def emit_one(_):
        return 1

    @solid
    def add(_, x, y):
        return x + y

    return emit_one, add


def test_execute_solid():
    emit_one, _ = get_solids()

    result = execute_in_process(emit_one, output_capturing_enabled=True)

    assert result.success
    assert result.output_values["result"] == 1


def test_execute_graph():
    emit_one, add = get_solids()

    @graph
    def emit_two():
        return add(emit_one(), emit_one())

    @graph
    def emit_three():
        return add(emit_two(), emit_one())

    result = execute_in_process(emit_three, output_capturing_enabled=True)

    assert result.success

    assert result.output_values["result"] == 3
    assert result.result_for_node("add").output_values["result"] == 3
    assert result.result_for_node("emit_two").output_values["result"] == 2
    assert result.result_for_node("emit_one").output_values["result"] == 1
    assert (
        result.result_for_node("emit_two").result_for_node("emit_one").output_values["result"] == 1
    )
    assert (
        result.result_for_node("emit_two").result_for_node("emit_one_2").output_values["result"]
        == 1
    )


def test_execute_solid_with_inputs():
    @solid
    def add_one(_, x):
        return 1 + x

    result = execute_in_process(add_one, input_values={"x": 5}, output_capturing_enabled=True)
    assert result.success

    assert result.output_values["result"] == 6


def test_execute_graph_with_inputs():
    emit_one, add = get_solids()

    @graph
    def add_one(x):
        return add(x, emit_one())

    result = execute_in_process(add_one, input_values={"x": 5}, output_capturing_enabled=True)
    assert result.success
    assert result.output_values["result"] == 6
    assert result.result_for_node("emit_one").output_values["result"] == 1


def test_execute_graph_nonexistent_inputs():
    emit_one, add = get_solids()

    @graph
    def get_two():
        return add(emit_one(), emit_one())

    with pytest.raises(
        DagsterInvalidDefinitionError,
        match=re.escape(
            'Invalid dependencies: graph "get_two" does not have input "x". Available inputs: []'
        ),
    ):
        execute_in_process(get_two, input_values={"x": 5})

    @graph
    def add_one(x):
        return add(x, emit_one())

    with pytest.raises(
        DagsterInvalidDefinitionError,
        match=re.escape(
            'Invalid dependencies: graph "add_one" does not have input "y". '
            "Available inputs: ['x']"
        ),
    ):
        execute_in_process(add_one, input_values={"y": 5})


def test_execute_solid_nonexistent_inputs():
    emit_one, add = get_solids()

    with pytest.raises(
        DagsterInvalidDefinitionError,
        match=re.escape(
            'Invalid dependencies: solid "emit_one" does not have input "x". Available inputs: []'
        ),
    ):
        execute_in_process(emit_one, input_values={"x": 5})

    with pytest.raises(
        DagsterInvalidDefinitionError,
        match=re.escape(
            'Invalid dependencies: solid "add" does not have input "z". '
            "Available inputs: ['x', 'y']"
        ),
    ):
        execute_in_process(add, input_values={"z": 5})


def test_dynamic_output_solid():
    @solid(output_defs=[DynamicOutputDefinition()])
    def should_work(_):
        yield DynamicOutput(1, mapping_key="1")
        yield DynamicOutput(2, mapping_key="2")

    result = execute_in_process(should_work)
    assert result.success
    assert result.output_values["result"]["1"] == 1
    assert result.output_values["result"]["2"] == 2



from __future__ import unicode_literals

import pytest
from dagster import pipeline
from dagster.core.errors import DagsterInvariantViolationError
from dagster.core.execution.plan.plan import ExecutionPlan
from dagster.serdes import deserialize_json_to_dagster_namedtuple

OLD_EXECUTION_PLAN_SNAPSHOT = """{
  "__class__": "ExecutionPlanSnapshot",
  "artifacts_persisted": true,
  "initial_known_state": null,
  "pipeline_snapshot_id": "0baebbaa257331ddeaccd3076c3dca413e099cf4",
  "step_keys_to_execute": [
    "solid_one",
    "solid_two"
  ],
  "steps": [
    {
      "__class__": "ExecutionStepSnap",
      "inputs": [],
      "key": "solid_one",
      "kind": {
        "__enum__": "StepKind.COMPUTE"
      },
      "metadata_items": [],
      "outputs": [
        {
          "__class__": "ExecutionStepOutputSnap",
          "dagster_type_key": "Any",
          "name": "result",
          "properties": {
            "__class__": "StepOutputProperties",
            "is_asset": false,
            "is_dynamic": false,
            "is_required": true,
            "should_materialize": false
          },
          "solid_handle": {
            "__class__": "SolidHandle",
            "name": "solid_one",
            "parent": null
          }
        }
      ],
      "solid_handle_id": "solid_one",
      "step_handle": {
        "__class__": "StepHandle",
        "solid_handle": {
          "__class__": "SolidHandle",
          "name": "solid_one",
          "parent": null
        }
      },
      "tags": {}
    },
    {
      "__class__": "ExecutionStepSnap",
      "inputs": [
        {
          "__class__": "ExecutionStepInputSnap",
          "dagster_type_key": "Any",
          "name": "num",
          "source": {
            "__class__": "FromStepOutput",
            "fan_in": false,
            "input_name": "num",
            "solid_handle": {
              "__class__": "SolidHandle",
              "name": "solid_two",
              "parent": null
            },
            "step_output_handle": {
              "__class__": "StepOutputHandle",
              "mapping_key": null,
              "output_name": "result",
              "step_key": "solid_one"
            }
          },
          "upstream_output_handles": [
            {
              "__class__": "StepOutputHandle",
              "mapping_key": null,
              "output_name": "result",
              "step_key": "solid_one"
            }
          ]
        }
      ],
      "key": "solid_two",
      "kind": {
        "__enum__": "StepKind.COMPUTE"
      },
      "metadata_items": [],
      "outputs": [
        {
          "__class__": "ExecutionStepOutputSnap",
          "dagster_type_key": "Any",
          "name": "result",
          "properties": {
            "__class__": "StepOutputProperties",
            "is_asset": false,
            "is_dynamic": false,
            "is_required": true,
            "should_materialize": false
          },
          "solid_handle": {
            "__class__": "SolidHandle",
            "name": "solid_two",
            "parent": null
          }
        }
      ],
      "solid_handle_id": "solid_two",
      "step_handle": {
        "__class__": "StepHandle",
        "solid_handle": {
          "__class__": "SolidHandle",
          "name": "solid_two",
          "parent": null
        }
      },
      "tags": {}
    }
  ]
}"""


@pipeline
def noop_pipeline():
    pass


def test_cant_load_old_snapshot():
    snapshot = deserialize_json_to_dagster_namedtuple(OLD_EXECUTION_PLAN_SNAPSHOT)
    with pytest.raises(
        DagsterInvariantViolationError,
        match="Tried to reconstruct an old ExecutionPlanSnapshot that was created before snapshots had enough information to fully reconstruct the ExecutionPlan",
    ):
        ExecutionPlan.rebuild_from_snapshot("noop_pipeline", snapshot)



from dagster import lambda_solid, pipeline, repository


@lambda_solid
def do_something():
    return 1


@pipeline(name="extra")
def extra_pipeline():
    do_something()


@repository
def extra():
    return {
        "pipelines": {"extra": extra_pipeline},
    }



import inspect
import sys
from typing import (
    AbstractSet,
    Any,
    Callable,
    Dict,
    Generator,
    List,
    NoReturn,
    Optional,
    Set,
    Tuple,
    Union,
)

Type = Union[type, Tuple[type, ...]]
Numeric = Union[int, float]


class CheckError(Exception):
    pass


class ParameterCheckError(CheckError):
    pass


class ElementCheckError(CheckError):
    pass


class NotImplementedCheckError(CheckError):
    pass


def _param_type_mismatch_exception(
    obj: Any, ttype: Type, param_name: str, additional_message: str = None
) -> ParameterCheckError:
    if isinstance(ttype, tuple):
        type_names = sorted([t.__name__ for t in ttype])
        return ParameterCheckError(
            'Param "{name}" is not one of {type_names}. Got {obj} which is type {obj_type}.'
            "{additional_message}".format(
                name=param_name,
                obj=repr(obj),
                type_names=type_names,
                obj_type=type(obj),
                additional_message=" " + additional_message if additional_message else "",
            )
        )
    else:
        return ParameterCheckError(
            'Param "{name}" is not a {type}. Got {obj} which is type {obj_type}.'
            "{additional_message}".format(
                name=param_name,
                obj=repr(obj),
                type=ttype.__name__,
                obj_type=type(obj),
                additional_message=" " + additional_message if additional_message else "",
            )
        )


def _not_type_param_subclass_mismatch_exception(obj: Any, param_name: str) -> ParameterCheckError:
    return ParameterCheckError(
        'Param "{name}" was supposed to be a type. Got {obj} of type {obj_type}'.format(
            name=param_name, obj=repr(obj), obj_type=type(obj)
        )
    )


def _param_subclass_mismatch_exception(
    obj: Any, superclass: Type, param_name: str
) -> ParameterCheckError:
    return ParameterCheckError(
        'Param "{name}" is a type but not a subclass of {superclass}. Got {obj} instead'.format(
            name=param_name, superclass=superclass, obj=obj
        )
    )


def _type_mismatch_error(obj: Any, ttype: Type, desc: str = None) -> CheckError:
    type_message = (
        f"not one of {sorted([t.__name__ for t in ttype])}"
        if isinstance(ttype, tuple)
        else f"not a {ttype.__name__}"
    )
    repr_obj = repr(obj)
    desc_str = f" Desc: {desc}" if desc else ""
    return CheckError(
        f"Object {repr_obj} is {type_message}. Got {repr_obj} with type {type(obj)}.{desc_str}"
    )


def _not_callable_exception(obj: Any, param_name: str) -> ParameterCheckError:
    return ParameterCheckError(
        'Param "{name}" is not callable. Got {obj} with type {obj_type}.'.format(
            name=param_name, obj=repr(obj), obj_type=type(obj)
        )
    )


def _param_invariant_exception(param_name: str, desc: str = None) -> ParameterCheckError:
    return ParameterCheckError(
        "Invariant violation for parameter {param_name}. Description: {desc}".format(
            param_name=param_name, desc=desc
        )
    )


def failed(desc: str) -> NoReturn:  # type: ignore[misc]
    if not isinstance(desc, str):
        raise CheckError("desc argument must be a string")

    raise CheckError("Failure condition: {desc}".format(desc=desc))


def not_implemented(desc: str):
    if not isinstance(desc, str):
        raise CheckError("desc argument must be a string")

    raise NotImplementedCheckError(f"Not implemented: {desc}")


def inst(obj: Any, ttype: Type, desc: str = None) -> Any:
    if not isinstance(obj, ttype):
        raise _type_mismatch_error(obj, ttype, desc)
    return obj


def opt_inst(obj: Any, ttype: Type, desc: str = None, default: Any = None) -> Any:
    if obj is not None and not isinstance(obj, ttype):
        raise _type_mismatch_error(obj, ttype, desc)
    return default if obj is None else obj


def subclass(obj: Any, superclass: Type, desc: str = None) -> Any:
    if not issubclass(obj, superclass):
        raise _type_mismatch_error(obj, superclass, desc)

    return obj


def is_callable(obj: Any, desc: str = None) -> Callable:
    if not callable(obj):
        if desc:
            raise CheckError(
                "Must be callable. Got {obj}. Description: {desc}".format(obj=repr(obj), desc=desc)
            )
        else:
            raise CheckError(
                "Must be callable. Got {obj}. Description: {desc}".format(obj=obj, desc=desc)
            )
    return obj


def not_none_param(obj: Any, param_name: str) -> Any:
    if obj is None:
        raise _param_invariant_exception(param_name, f"Param {param_name} cannot be none")
    return obj


def invariant(condition: Any, desc: str = None) -> bool:
    if not condition:
        if desc:
            raise CheckError(f"Invariant failed. Description: {desc}")
        else:
            raise CheckError("Invariant failed.")

    return True


def param_invariant(condition: Any, param_name: str, desc: str = None):
    if not condition:
        raise _param_invariant_exception(param_name, desc)


def inst_param(obj: Any, param_name: str, ttype: Type, additional_message: str = None) -> Any:
    if not isinstance(obj, ttype):
        raise _param_type_mismatch_exception(
            obj, ttype, param_name, additional_message=additional_message
        )
    return obj


def opt_inst_param(obj: Any, param_name: str, ttype: Type, default: Any = None) -> Any:
    if obj is not None and not isinstance(obj, ttype):
        raise _param_type_mismatch_exception(obj, ttype, param_name)
    return default if obj is None else obj


def callable_param(obj: Any, param_name: str) -> Callable:
    if not callable(obj):
        raise _not_callable_exception(obj, param_name)
    return obj


def opt_callable_param(obj: Any, param_name: str, default: Callable = None) -> Optional[Callable]:
    if obj is not None and not callable(obj):
        raise _not_callable_exception(obj, param_name)
    return default if obj is None else obj


def int_param(obj: Any, param_name: str) -> int:
    if not isinstance(obj, int):
        raise _param_type_mismatch_exception(obj, int, param_name)
    return obj


def int_value_param(obj: Any, value: int, param_name: str) -> int:
    if not isinstance(obj, int):
        raise _param_type_mismatch_exception(obj, int, param_name)
    if obj != value:
        raise _param_invariant_exception(param_name, f"Should be equal to {value}")

    return obj


def opt_int_param(obj: Any, param_name: str, default: int = None) -> Optional[int]:
    if obj is not None and not isinstance(obj, int):
        raise _param_type_mismatch_exception(obj, int, param_name)
    return default if obj is None else obj


def float_param(obj: Any, param_name: str) -> float:
    if not isinstance(obj, float):
        raise _param_type_mismatch_exception(obj, float, param_name)
    return obj


def opt_numeric_param(obj: Any, param_name: str, default: Numeric = None) -> Optional[Numeric]:
    if obj is not None and not isinstance(obj, (int, float)):
        raise _param_type_mismatch_exception(obj, (int, float), param_name)
    return default if obj is None else obj


def numeric_param(obj: Any, param_name: str) -> Numeric:
    if not isinstance(obj, (int, float)):
        raise _param_type_mismatch_exception(obj, (int, float), param_name)
    return obj


def opt_float_param(obj: Any, param_name: str, default: float = None) -> Optional[float]:
    if obj is not None and not isinstance(obj, float):
        raise _param_type_mismatch_exception(obj, float, param_name)
    return default if obj is None else obj


def str_param(obj: Any, param_name: str) -> str:
    if not isinstance(obj, str):
        raise _param_type_mismatch_exception(obj, str, param_name)
    return obj


def opt_str_param(obj: Any, param_name: str, default: str = None) -> Optional[str]:
    if obj is not None and not isinstance(obj, str):
        raise _param_type_mismatch_exception(obj, str, param_name)
    return default if obj is None else obj


def opt_nonempty_str_param(obj: Any, param_name: str, default: str = None) -> Optional[str]:
    if obj is not None and not isinstance(obj, str):
        raise _param_type_mismatch_exception(obj, str, param_name)
    return default if obj is None or obj == "" else obj


def bool_param(obj: Any, param_name: str) -> bool:
    if not isinstance(obj, bool):
        raise _param_type_mismatch_exception(obj, bool, param_name)
    return obj


def opt_bool_param(obj: Any, param_name: str, default: bool = None) -> Optional[bool]:
    if obj is not None and not isinstance(obj, bool):
        raise _param_type_mismatch_exception(obj, bool, param_name)
    return default if obj is None else obj


def is_list(obj_list: Any, of_type: Type = None, desc: str = None) -> List:
    if not isinstance(obj_list, list):
        raise _type_mismatch_error(obj_list, list, desc)

    if not of_type:
        return obj_list

    return _check_list_items(obj_list, of_type)


def is_tuple(obj_tuple: Any, of_type: Type = None, desc: str = None) -> Tuple:
    if not isinstance(obj_tuple, tuple):
        raise _type_mismatch_error(obj_tuple, tuple, desc)

    if not of_type:
        return obj_tuple

    return _check_tuple_items(obj_tuple, of_type)


def list_param(obj_list: Any, param_name: str, of_type: Type = None) -> List:
    from dagster.utils import frozenlist

    if not isinstance(obj_list, (frozenlist, list)):
        raise _param_type_mismatch_exception(obj_list, (frozenlist, list), param_name)

    if not of_type:
        return obj_list

    return _check_list_items(obj_list, of_type)


def set_param(obj_set: Any, param_name: str, of_type: Type = None) -> AbstractSet:
    if not isinstance(obj_set, (frozenset, set)):
        raise _param_type_mismatch_exception(obj_set, (frozenset, set), param_name)

    if not of_type:
        return obj_set

    return _check_set_items(obj_set, of_type)


def tuple_param(obj: Any, param_name: str, of_type: Type = None) -> Tuple:
    if not isinstance(obj, tuple):
        raise _param_type_mismatch_exception(obj, tuple, param_name)

    if of_type is None:
        return obj

    return _check_tuple_items(obj, of_type)


def matrix_param(matrix: Any, param_name: str, of_type: Type = None) -> List[List]:
    matrix = list_param(matrix, param_name, of_type=list)
    if not matrix:
        raise CheckError("You must pass a list of lists. Received an empty list.")
    for sublist in matrix:
        sublist = list_param(sublist, "sublist_{}".format(param_name), of_type=of_type)
        if len(sublist) != len(matrix[0]):
            raise CheckError("All sublists in matrix must have the same length")
    return matrix


def opt_tuple_param(
    obj: Any, param_name: str, default: Tuple = None, of_type: Type = None
) -> Optional[Tuple]:
    if obj is not None and not isinstance(obj, tuple):
        raise _param_type_mismatch_exception(obj, tuple, param_name)

    if obj is None:
        return default

    if of_type is None:
        return obj

    return _check_tuple_items(obj, of_type)


def _check_list_items(obj_list: Any, of_type: Type) -> List:
    for obj in obj_list:
        if not isinstance(obj, of_type):
            if isinstance(obj, type):
                additional_message = (
                    " Did you pass a class where you were expecting an instance of the class?"
                )
            else:
                additional_message = ""
            raise CheckError(
                "Member of list mismatches type. Expected {of_type}. Got {obj_repr} of type "
                "{obj_type}.{additional_message}".format(
                    of_type=of_type,
                    obj_repr=repr(obj),
                    obj_type=type(obj),
                    additional_message=additional_message,
                )
            )

    return obj_list


def _check_set_items(obj_set: Any, of_type: Type) -> Set:
    for obj in obj_set:

        if not isinstance(obj, of_type):
            if isinstance(obj, type):
                additional_message = (
                    " Did you pass a class where you were expecting an instance of the class?"
                )
            else:
                additional_message = ""
            raise CheckError(
                "Member of set mismatches type. Expected {of_type}. Got {obj_repr} of type "
                "{obj_type}.{additional_message}".format(
                    of_type=of_type,
                    obj_repr=repr(obj),
                    obj_type=type(obj),
                    additional_message=additional_message,
                )
            )

    return obj_set


def _check_tuple_items(obj_tuple: Any, of_type: Union[Tuple, Type]) -> Tuple:
    if isinstance(of_type, tuple):
        len_tuple = len(obj_tuple)
        len_type = len(of_type)
        if not len_tuple == len_type:
            raise CheckError(
                "Tuple mismatches type: tuple had {len_tuple} members but type had "
                "{len_type}".format(len_tuple=len_tuple, len_type=len_type)
            )

        for (i, obj) in enumerate(obj_tuple):
            of_type_i = of_type[i]
            if not isinstance(obj, of_type_i):
                if isinstance(obj, type):
                    additional_message = (
                        " Did you pass a class where you were expecting an instance of the class?"
                    )
                else:
                    additional_message = ""
                raise CheckError(
                    "Member of tuple mismatches type at index {index}. Expected {of_type}. Got "
                    "{obj_repr} of type {obj_type}.{additional_message}".format(
                        index=i,
                        of_type=of_type_i,
                        obj_repr=repr(obj),
                        obj_type=type(obj),
                        additional_message=additional_message,
                    )
                )

    else:
        for (i, obj) in enumerate(obj_tuple):
            if not isinstance(obj, of_type):
                if isinstance(obj, type):
                    additional_message = (
                        " Did you pass a class where you were expecting an instance of the class?"
                    )
                else:
                    additional_message = ""
                raise CheckError(
                    "Member of tuple mismatches type at index {index}. Expected {of_type}. Got "
                    "{obj_repr} of type {obj_type}.{additional_message}".format(
                        index=i,
                        of_type=of_type,
                        obj_repr=repr(obj),
                        obj_type=type(obj),
                        additional_message=additional_message,
                    )
                )

    return obj_tuple


def opt_list_param(obj_list: Any, param_name: str, of_type: Type = None) -> List:
    """Ensures argument obj_list is a list or None; in the latter case, instantiates an empty list
    and returns it.

    If the of_type argument is provided, also ensures that list items conform to the type specified
    by of_type.
    """
    from dagster.utils import frozenlist

    if obj_list is not None and not isinstance(obj_list, (frozenlist, list)):
        raise _param_type_mismatch_exception(obj_list, (frozenlist, list), param_name)

    if not obj_list:
        return []
    if not of_type:
        return obj_list

    return _check_list_items(obj_list, of_type)


def opt_set_param(obj_set: Any, param_name: str, of_type: Type = None) -> AbstractSet:
    """Ensures argument obj_set is a set or None; in the latter case, instantiates an empty set
    and returns it.

    If the of_type argument is provided, also ensures that list items conform to the type specified
    by of_type.
    """
    if obj_set is not None and not isinstance(obj_set, (frozenset, set)):
        raise _param_type_mismatch_exception(obj_set, (frozenset, set), param_name)
    if not obj_set:
        return set()
    if not of_type:
        return obj_set

    return _check_set_items(obj_set, of_type)


def opt_nullable_list_param(obj_list: Any, param_name: str, of_type: Type = None) -> Optional[List]:
    """Ensures argument obj_list is a list or None. Returns None if input is None.

    If the of_type argument is provided, also ensures that list items conform to the type specified
    by of_type.
    """
    from dagster.utils import frozenlist

    if obj_list is not None and not isinstance(obj_list, (frozenlist, list)):
        raise _param_type_mismatch_exception(obj_list, (frozenlist, list), param_name)

    if not obj_list:
        return None if obj_list is None else []
    if not of_type:
        return obj_list

    return _check_list_items(obj_list, of_type)


def _check_key_value_types(
    obj: Any,
    key_type: Type = None,
    value_type: Type = None,
    key_check: Callable = isinstance,
    value_check: Callable = isinstance,
) -> Dict:
    """Ensures argument obj is a dictionary, and enforces that the keys/values conform to the types
    specified by key_type, value_type.
    """
    if not isinstance(obj, dict):
        raise _type_mismatch_error(obj, dict)

    for key, value in obj.items():
        if key_type and not key_check(key, key_type):
            raise CheckError(
                "Key in dictionary mismatches type. Expected {key_type}. Got {obj_repr}".format(
                    key_type=repr(key_type), obj_repr=repr(key)
                )
            )

        if value_type and not value_check(value, value_type):
            raise CheckError(
                "Value in dictionary mismatches expected type for key {key}. Expected value "
                "of type {vtype}. Got value {value} of type {obj_type}.".format(
                    vtype=repr(value_type), obj_type=type(value), key=key, value=value
                )
            )

    return obj


def dict_param(
    obj: Any,
    param_name: str,
    key_type: Type = None,
    value_type: Type = None,
    additional_message: str = None,
) -> Dict:
    """Ensures argument obj is a native Python dictionary, raises an exception if not, and otherwise
    returns obj.
    """
    from dagster.utils import frozendict

    if not isinstance(obj, (frozendict, dict)):
        raise _param_type_mismatch_exception(
            obj, (frozendict, dict), param_name, additional_message=additional_message
        )

    if not (key_type or value_type):
        return obj

    return _check_key_value_types(obj, key_type, value_type)


def opt_dict_param(
    obj: Any,
    param_name: str,
    key_type: Type = None,
    value_type: Type = None,
    value_class: Type = None,
) -> Dict:
    """Ensures argument obj is either a dictionary or None; if the latter, instantiates an empty
    dictionary.
    """
    from dagster.utils import frozendict

    if obj is not None and not isinstance(obj, (frozendict, dict)):
        raise _param_type_mismatch_exception(obj, (frozendict, dict), param_name)

    if not obj:
        return {}

    if value_class:
        return _check_key_value_types(obj, key_type, value_type=value_class, value_check=issubclass)
    return _check_key_value_types(obj, key_type, value_type)


def opt_nullable_dict_param(
    obj: Any,
    param_name: str,
    key_type: Type = None,
    value_type: Type = None,
    value_class: Type = None,
) -> Optional[Dict]:
    """Ensures argument obj is either a dictionary or None;"""
    from dagster.utils import frozendict

    if obj is not None and not isinstance(obj, (frozendict, dict)):
        raise _param_type_mismatch_exception(obj, (frozendict, dict), param_name)

    if not obj:
        return None if obj is None else {}

    if value_class:
        return _check_key_value_types(obj, key_type, value_type=value_class, value_check=issubclass)
    return _check_key_value_types(obj, key_type, value_type)


def _check_two_dim_key_value_types(
    obj: Any, key_type: Type = None, _param_name: str = None, value_type: Type = None
) -> Dict:
    _check_key_value_types(obj, key_type, dict)  # check level one

    for level_two_dict in obj.values():
        _check_key_value_types(level_two_dict, key_type, value_type)  # check level two

    return obj


def two_dim_dict_param(
    obj: Any, param_name: str, key_type: Type = str, value_type: Type = None
) -> Dict:
    if not isinstance(obj, dict):
        raise _param_type_mismatch_exception(obj, dict, param_name)

    return _check_two_dim_key_value_types(obj, key_type, param_name, value_type)


def opt_two_dim_dict_param(
    obj: Any, param_name: str, key_type: Type = str, value_type: Type = None
) -> Dict:
    if obj is not None and not isinstance(obj, dict):
        raise _param_type_mismatch_exception(obj, dict, param_name)

    if not obj:
        return {}

    return _check_two_dim_key_value_types(obj, key_type, param_name, value_type)


def type_param(obj: Any, param_name: str) -> type:
    if not isinstance(obj, type):
        raise _not_type_param_subclass_mismatch_exception(obj, param_name)
    return obj


def opt_type_param(obj: Any, param_name: str, default: type = None) -> Optional[type]:
    if obj is not None and not isinstance(obj, type):
        raise _not_type_param_subclass_mismatch_exception(obj, param_name)
    return obj if obj is not None else default


def subclass_param(obj: Any, param_name: str, superclass: type) -> type:
    type_param(obj, param_name)
    if not issubclass(obj, superclass):
        raise _param_subclass_mismatch_exception(obj, superclass, param_name)

    return obj


def opt_subclass_param(obj: Any, param_name: str, superclass: type) -> Optional[type]:
    opt_type_param(obj, param_name)
    if obj is not None and not issubclass(obj, superclass):
        raise _param_subclass_mismatch_exception(obj, superclass, param_name)

    return obj


def _element_check_error(key: Any, value: Any, ddict: Dict, ttype: Type) -> ElementCheckError:
    return ElementCheckError(
        "Value {value} from key {key} is not a {ttype}. Dict: {ddict}".format(
            key=key, value=repr(value), ddict=repr(ddict), ttype=repr(ttype)
        )
    )


def generator(obj: Any) -> Generator:
    if not inspect.isgenerator(obj):
        raise ParameterCheckError(
            f"Not a generator (return value of function that yields) Got {obj} instead"
        )
    return obj


def opt_generator(obj: Any) -> Optional[Generator]:
    if obj is not None and not inspect.isgenerator(obj):
        raise ParameterCheckError(
            f"Not a generator (return value of function that yields) Got {obj} instead"
        )
    return obj


def generator_param(obj: Any, param_name: str) -> Generator:
    if not inspect.isgenerator(obj):
        raise ParameterCheckError(
            f'Param "{param_name}" is not a generator (return value of function that yields) Got '
            f"{obj} instead"
        )
    return obj


def opt_generator_param(obj: Any, param_name: str) -> Optional[Generator]:
    if obj is not None and not inspect.isgenerator(obj):
        raise ParameterCheckError(
            f'Param "{param_name}" is not a generator (return value of function that yields) Got '
            f"{obj} instead"
        )
    return obj


def list_elem(ddict: Dict, key: str, of_type: Type = None) -> List:  # type: ignore[return]
    dict_param(ddict, "ddict")
    str_param(key, "key")
    opt_type_param(of_type, "of_type")

    value = ddict.get(key)

    if isinstance(value, list):
        if not of_type:
            return value

        return _check_list_items(value, of_type)

    raise _element_check_error(key, value, ddict, list)


def opt_list_elem(ddict: Dict, key: str, of_type: Type = None) -> List:
    dict_param(ddict, "ddict")
    str_param(key, "key")
    opt_type_param(of_type, "of_type")

    value = ddict.get(key)

    if value is None:
        return []

    if not isinstance(value, list):
        raise _element_check_error(key, value, ddict, list)

    if not of_type:
        return value

    return _check_list_items(value, of_type)


def dict_elem(ddict: Dict, key: str) -> Dict:
    from dagster.utils import frozendict

    dict_param(ddict, "ddict")
    str_param(key, "key")

    if key not in ddict:
        raise CheckError(f"{key} not present in dictionary {ddict}")

    value = ddict[key]
    if not isinstance(value, (frozendict, dict)):
        raise _element_check_error(key, value, ddict, (frozendict, dict))
    return value


def opt_dict_elem(ddict: Dict, key: str) -> Dict:
    from dagster.utils import frozendict

    dict_param(ddict, "ddict")
    str_param(key, "key")

    value = ddict.get(key)

    if value is None:
        return {}

    if not isinstance(value, (frozendict, dict)):
        raise _element_check_error(key, value, ddict, list)

    return value


def bool_elem(ddict: Dict, key: str) -> bool:
    dict_param(ddict, "ddict")
    str_param(key, "key")

    value = ddict[key]
    if not isinstance(value, bool):
        raise _element_check_error(key, value, ddict, bool)
    return value


def opt_float_elem(ddict: Dict, key: str) -> Optional[float]:
    dict_param(ddict, "ddict")
    str_param(key, "key")

    value = ddict.get(key)
    if value is None:
        return None
    if not isinstance(value, float):
        raise _element_check_error(key, value, ddict, float)
    return value


def float_elem(ddict: Dict, key: str) -> float:
    dict_param(ddict, "ddict")
    str_param(key, "key")

    value = ddict[key]
    if not isinstance(value, float):
        raise _element_check_error(key, value, ddict, float)
    return value


def opt_int_elem(ddict: Dict, key: str) -> Optional[int]:
    dict_param(ddict, "ddict")
    str_param(key, "key")

    value = ddict.get(key)
    if value is None:
        return None
    if not isinstance(value, int):
        raise _element_check_error(key, value, ddict, int)
    return value


def int_elem(ddict: Dict, key: str) -> int:
    dict_param(ddict, "ddict")
    str_param(key, "key")

    value = ddict[key]
    if not isinstance(value, int):
        raise _element_check_error(key, value, ddict, int)
    return value


def opt_str_elem(ddict: Dict, key: str) -> Optional[str]:
    dict_param(ddict, "ddict")
    str_param(key, "key")

    value = ddict.get(key)
    if value is None:
        return None
    if not isinstance(value, str):
        raise _element_check_error(key, value, ddict, str)
    return value


def str_elem(ddict: Dict, key: str) -> str:
    dict_param(ddict, "ddict")
    str_param(key, "key")

    value = ddict[key]
    if not isinstance(value, str):
        raise _element_check_error(key, value, ddict, str)
    return value


def class_param(obj: Any, param_name: str) -> Union[ParameterCheckError, type]:
    if not inspect.isclass(obj):
        return ParameterCheckError(
            f'Param "{param_name}" is not a class. Got {repr(obj)} which is type {type(obj)}.'
        )
    return obj
